/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998,1999  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * Version: 1.1 04/12/1999 Maxim Krasnyansky <max_mk@yahoo.com>
 */ 
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <syslog.h>
#include <strings.h>

#include <sys/time.h>
#include <sys/resource.h>
#include <sched.h>
#include <errno.h>

#include "vtun.h"
#include "linkfd.h"
#include "lib.h"

struct lfd_mod *lfd_mod_head = NULL, *lfd_mod_tail = NULL;

extern int errno;

/* Modules functions*/

/* Add module to the end of modules list */
void inline lfd_add_mod(struct lfd_mod *mod)
{
     if( !lfd_mod_head ){
        lfd_mod_head = lfd_mod_tail = mod;
	mod->next = mod->prev = NULL;
     } else {
        lfd_mod_tail->next = mod;
        mod->prev = lfd_mod_tail;
        mod->next = NULL;
        lfd_mod_tail = mod;
     }
}

/*  Initialize and allocate each module */
int inline lfd_alloc_mod(struct vtun_host *host)
{
     struct lfd_mod *mod = lfd_mod_head;

     while( mod ){
        if( mod->alloc && (mod->alloc)(host) )
	   return 1; 
	mod = mod->next;
     } 

     return 0;
}

/* Free each module */
int inline lfd_free_mod(void)
{
     struct lfd_mod *mod = lfd_mod_head;

     while( mod ){
        if( mod->free && (mod->free)() )
	   return 1;
	mod = mod->next;
     } 

     return 0;
}

 /* Run modules down (from head to tail) */
int inline lfd_run_down(int len, char *in, char **out)
{
     struct lfd_mod *mod = lfd_mod_head;
     
     if( !mod )
	*out = in;
     else		   	
        for( ; mod && len > 0; mod = mod->next )
            if( mod->encode ){
               len = (mod->encode)(len,in,out);
               in = *out;
            }
     return len;
}

/* Run modules up (from tail to head) */
int inline lfd_run_up(int len, char *in, char **out)
{
     struct lfd_mod *mod = lfd_mod_tail;
     
     if( !mod )
	*out = in;
     else		   	
        for( ; mod && len > 0; mod = mod->prev )
	    if( mod->decode ){
	       len = (mod->decode)(len,in,out);
               in = *out;
	    }
     return len;
}

/* Check if modules are accepting the data(down) */
int inline lfd_check_down(void)
{
     struct lfd_mod *mod = lfd_mod_head;
     int err = 1;	    
 
     if( mod ){
	for( ; mod && err > 0; mod = mod->next )
           if( mod->avail_encode )
              err = (mod->avail_encode)();
     }
	
     return err;
}

/* Check if modules are accepting the data(up) */
int inline lfd_check_up(void)
{
     struct lfd_mod *mod = lfd_mod_tail;
     int err = 1;	    

     if( mod ){
        for( ; mod && err > 0; mod = mod->prev)
           if( mod->avail_decode )
              err = (mod->avail_decode)();
     }

     return err;
}

/* Shaper */

int lfd_err = 0;
int lfd_len = 0; 
int lfd_fd1 = 0;
int lfd_fd2 = 0;
     
char *lfd_buf;

void lfd_shaper_timer(int sig){
     static struct timeval tv={0,0};
     static fd_set fdset;
     char *out = lfd_buf;
     register int len;

     FD_ZERO(&fdset);
     FD_SET(lfd_fd2, &fdset);

     /* 
      * Check for available data, or we will sleep in the read call 
      */
     if( (len = select(lfd_fd2+1, &fdset, NULL, NULL, &tv)) == -1) {
	lfd_err = 1;
	return;
     }
  
     if( !len )
	return;

     if( (len=read(lfd_fd2, lfd_buf, lfd_len)) <= 0 )
 	lfd_err = 1;
     else {
        if( (len = lfd_run_down(len,lfd_buf,&out)) == -1 )
	   lfd_err = 1;	
	else
   	   if( write_n(lfd_fd1, out, len) < 0 )
	      lfd_err = 1;
     }
}

int lfd_shaper(int fd1, int fd2, int speed){
     struct itimerval itm;
     struct sigaction sa;
     char *buf, *out;

     if(speed < 64)
	  return 1;
     /*
      * Calculate how much we will copy each timer tick. 
      * It is very primitive shaping, but shaping results 
      * are prety accurate.
      */
 
     lfd_len=LINKFD_MAGIC * (speed / 64);

     if( !(lfd_buf=malloc(lfd_len)) )
         return 1; 

     if( !(buf=malloc(LINKFD_BUF_SIZE)) )
        return 1; 
     out = buf;

     lfd_fd1 = fd1;
     lfd_fd2 = fd2;
     lfd_err = 0;

     /* Set up the SIGALRM handler */
     memset(&sa,0,sizeof(sa));    
     sa.sa_handler=lfd_shaper_timer;
     sigemptyset(&sa.sa_mask);
     sigfillset(&sa.sa_mask);
     sigaction(SIGALRM,&sa,NULL);    

     /* Start the timer */
     memset(&itm,0,sizeof(itm));
     itm.it_interval.tv_usec = itm.it_value.tv_usec = LINKFD_FREQ;
     setitimer(ITIMER_REAL,&itm,NULL);
     
     while(!lfd_err){
        register int len;

	if( (len=read(fd1, buf, LINKFD_BUF_SIZE)) <= 0 )
	   if( errno == EINTR )
	      continue;
	   else
	      break; 
        if( (len = lfd_run_up(len,buf,&out)) == -1 )
	   break;	 

	if( write_n(fd2, out, len) < 0 )
   	   break; 
     }

     memset(&itm,0,sizeof(itm));
     setitimer(ITIMER_REAL,&itm,NULL);

     sa.sa_handler=SIG_DFL;
     sa.sa_flags = 0;
     sigaction(SIGALRM,&sa,NULL);    

     free(lfd_buf);
     free(buf);
	
     return 0;
}
/* End of Shaper */

/* Linker */

int lfd_linker(int fd1,int fd2)
{
     char *buf, *out;
     fd_set fdset;
     int maxfd;

     if( !(buf=malloc(LINKFD_BUF_SIZE)) ){
	syslog(LOG_ERR,"Can't allocate linker buffer"); 
        return 1; 
     }

     maxfd = (fd1 > fd2 ? fd1 : fd2) + 1;

     while(1){
        register int len;

        /* Wait for data */
        FD_ZERO(&fdset);
	FD_SET(fd1, &fdset);
	FD_SET(fd2, &fdset);
	if( select(maxfd, &fdset, NULL, NULL, NULL) < 0 )
	   break;

	if( FD_ISSET(fd2, &fdset) && lfd_check_down() ){
	   if( (len=read(fd2, buf, LINKFD_BUF_SIZE)) <= 0 )
	      break;
	   if( (len=lfd_run_down(len,buf,&out)) == -1 )
	      break;	 
	   if( write_n(fd1, out, len) < 0 )
	      break;
	}
	if( FD_ISSET(fd1, &fdset) && lfd_check_up() ){
	   if( (len=read(fd1, buf, LINKFD_BUF_SIZE)) <= 0 )
	      break;  
	   if( (len=lfd_run_up(len,buf,&out)) == -1 )
	      break;	 
	   if( write_n(fd2, out, len) < 0 )
	      break; 
	}
     }
     free(buf);

     return 0;
}

/* Link remote and local file descriptors */ 

int linkfd(struct vtun_host *host)
{
     int old_prio;

     old_prio=getpriority(PRIO_PROCESS,0);
     setpriority(PRIO_PROCESS,0,LINKFD_PRIO);

     if(host->flags & VTUN_COMPRESS)
	lfd_add_mod(&lfd_zcomp);

     if(host->flags & VTUN_ENCRYPT)
	lfd_add_mod(&lfd_encrypt);

     if( lfd_alloc_mod(host) )
	return 1;

     if(host->flags & VTUN_SHAPE)
	lfd_shaper(host->rmt_fd, host->loc_fd, host->spd_out);
     else
        lfd_linker(host->rmt_fd, host->loc_fd);

     lfd_free_mod();

     setpriority(PRIO_PROCESS,0,old_prio);

     return 0;
}
